from ruleengine.dsl.condition import Condition
from ruleengine.dsl.base_conditions import topic, ts, concat
from ruleengine.dsl.action import Action
from typing import Optional


def noop_upload(
    trigger_ts: int,  # todo delete later
    before: int,
    title: str,
    description: str,
    labels: list[str],
    extra_files: list[str],
):
    pass


def noop_create_moment(
    title: str,
    description: str,
    timestamp: int,
    duration: int,
    create_task: bool,
    assign_to: Optional[str],
):
    pass


noop = {"upload": noop_upload, "create_moment": noop_create_moment}


class ForwardingAction(Action):
    def __init__(self, thunk, args):
        self.__thunk = thunk
        self.__args = args

    def run(self, item, scope):
        actual_args = {}
        for name, value in self.__args.items():
            if isinstance(value, Condition):
                new_value, _ = value.evaluate_condition_at(item, scope)
                actual_args[name] = new_value
            else:
                actual_args[name] = value
        self.__thunk(**actual_args)


def upload_factory(impl):
    def res(
        title=concat("Device auto upload @ ", ts),
        description="",
        labels=[],
        extra_files=[],
        before=10,
    ):
        args = {
            "before": before,
            "title": Condition.map(Condition.wrap(title), str),
            "description": Condition.map(Condition.wrap(description), str),
            "labels": labels,
            "extra_files": extra_files,
            "trigger_ts": ts,
        }

        return ForwardingAction(impl, args)

    return res


def create_moment_factory(impl):
    def res(
        title,
        description="",
        timestamp=ts,
        duration=1,
        create_task=False,
        assign_to=None,
    ):
        args = {
            "title": Condition.map(Condition.wrap(title), str),
            "description": Condition.map(Condition.wrap(description), str),
            "timestamp": Condition.map(Condition.wrap(timestamp), int),
            "duration": Condition.map(Condition.wrap(duration), int),
            "create_task": Condition.wrap(create_task),
            "assign_to": assign_to,
        }
        return ForwardingAction(impl, args)

    return res



from .base_conditions import and_
from .condition import Condition


def sustained(context_condition, variable_condition, duration):
    """
    This condition triggers when the variable condition continues to be true for
    the given duration.

    The context condition is used to limit the scope of the variable condition.
    For example, we might say "if topic X has value Y for 10 seconds", which
    translates to context being topic == X, and variable being value == Y
    """
    return and_(
        context_condition,
        RisingEdgeCondition(SustainedCondition(variable_condition, duration)),
    )


def sequential(*conditions, duration=None):
    return SequenceMatchCondition(list(conditions), duration)


def timeout(*conditions, duration):
    return SequenceMatchCondition(list(conditions), duration, True)


def any_order(*conditions):
    return RisingEdgeCondition(AnyOrderCondition(list(conditions)))


def repeated(condition, times, duration):
    assert times > 1, "In repeated condition, times must be more than 1"
    return and_(
        condition,
        RisingEdgeCondition(SequenceMatchCondition([condition] * times, duration)),
    )


def debounce(condition, duration):
    return and_(condition, RisingEdgeCondition(condition, duration))


def throttle(condition, duration):
    return ThrottleCondition(condition, duration)


class RisingEdgeCondition(Condition):
    """
    A condition that detects when child condition changes from false to true.

    Once child condition becomes true, this condition will not fire unless the
    child condition first becomes false, then back to true.

    If max_gap is given, and two invocations of this condition have timestamps
    further apart than the gap, the state is reset. It behaves as if a false is
    inserted in the gap.
    """

    def __init__(self, condition, max_gap=None):
        assert isinstance(condition, Condition)
        self.__condition = condition
        self.__active = False
        self.__last_activation = None
        self.__max_gap = max_gap

    def evaluate_condition_at(self, item, scope):
        value, new_scope = self.__condition.evaluate_condition_at(item, scope)
        if not value:
            self.__active = False
            return False, new_scope

        if self.__active and (
            self.__max_gap is None or item.ts - self.__last_activation < self.__max_gap
        ):
            self.__last_activation = item.ts
            return False, new_scope

        self.__active = True
        self.__last_activation = item.ts
        return True, new_scope


class SustainedCondition(Condition):
    def __init__(self, condition, duration=-1):
        super().__init__()

        assert isinstance(condition, Condition)
        self.__condition = condition
        self.__duration = duration
        self.__start = None

    def evaluate_condition_at(self, item, scope):
        value, new_scope = self.__condition.evaluate_condition_at(item, scope)
        if not value:
            self.__start = None
            return False, new_scope

        if self.__start is None:
            self.__start = item.ts

        if item.ts - self.__start > self.__duration:
            return True, {**new_scope, "start_time": self.__start}

        return False, new_scope


class SequenceMatchCondition(Condition):
    def __init__(self, sequence, duration=None, trigger_on_timeout=False):
        super().__init__()

        assert len(sequence) > 1, "Sequence must be longer than 1"
        for c in sequence:
            assert isinstance(c, Condition)

        self.__seq = sequence
        self.__duration = duration
        self.__trigger_on_timeout = trigger_on_timeout

        # List of (start time, current index, scope) tuple. Each item denotes
        # one ongoing matching sequence, with start time being the time of the
        # first matched condition, and current index is the next condition to be
        # matched
        self.__ongoings = []

    def evaluate_condition_at(self, item, scope):
        ongoings = []
        success = None
        for start, curr_index, curr_scope in self.__ongoings:
            if self.__duration is not None and item.ts - start > self.__duration:
                if self.__trigger_on_timeout:
                    success = start, curr_scope
                continue

            matched, new_scope = self.__seq[curr_index].evaluate_condition_at(
                item, curr_scope
            )
            if matched:
                if curr_index + 1 == len(self.__seq):
                    if not self.__trigger_on_timeout:
                        success = start, new_scope
                else:
                    ongoings.append((start, curr_index + 1, new_scope))

            else:
                ongoings.append((start, curr_index, curr_scope))

        self.__ongoings = ongoings

        matched, new_scope = self.__seq[0].evaluate_condition_at(item, scope)
        if matched:
            self.__ongoings.append((item.ts, 1, new_scope))

        if success:
            return True, {**success[1], "start_time": success[0]}

        return False, scope


class ThrottleCondition(Condition):
    def __init__(self, condition, duration):
        assert isinstance(condition, Condition)
        self.__condition = condition
        self.__duration = duration
        self.__last_trigger = -duration

    def evaluate_condition_at(self, item, scope):
        if item.ts - self.__last_trigger < self.__duration:
            return False, scope

        value, scope = self.__condition.evaluate_condition_at(item, scope)
        if value:
            self.__last_trigger = item.ts
            return value, scope

        return False, scope


class AnyOrderCondition(Condition):
    def __init__(self, conditions):
        self.__unsatisfied = conditions
        self.__curr_scope = None

    def evaluate_condition_at(self, item, scope):
        if self.__curr_scope is None:
            self.__curr_scope = scope

        new_conditions = []
        for c in self.__unsatisfied:
            value, new_scope = c.evaluate_condition_at(item, self.__curr_scope)
            if value:
                self.__curr_scope = new_scope
            else:
                new_conditions.append(c)

        if not new_conditions:
            return True, self.__curr_scope

        self.__unsatisfied = new_conditions
        return False, self.__curr_scope


__all__ = [
    "any_order",
    "debounce",
    "repeated",
    "sequential",
    "sustained",
    "throttle",
    "timeout",
]



import unittest
from collections import namedtuple

from ruleengine.dsl.action import Action
from ruleengine.dsl.base_conditions import *
from ruleengine.engine import Engine, Rule, DiagnosisItem

MockMessage = namedtuple("MockMessage", "int_value str_value")

simple_sequence = [
    DiagnosisItem("t1", MockMessage(1, "hello"), 0, "MockMessage"),
    DiagnosisItem("t2", MockMessage(2, "hello"), 1, "MockMessage"),
    DiagnosisItem("t1", MockMessage(3, "heLlo"), 2, "MockMessage"),
    DiagnosisItem("t2", MockMessage(4, "hello"), 3, "MockMessage"),
    DiagnosisItem("t2", MockMessage(5, "world"), 4, "MockMessage"),
    DiagnosisItem(
        "t3",
        MockMessage(5, "The value is 324, which is expected to be less than 22"),
        5,
        "MockMessage",
    ),
    DiagnosisItem(
        "t3",
        MockMessage(5, "The value is 11, which is expected to be less than 22"),
        5,
        "MockMessage",
    ),
]


class CollectAction(Action):
    def __init__(self):
        self.collector = []

    def run(self, item, scope):
        self.collector.append(item)


class BaseConditionTest(unittest.TestCase):
    def test_always(self):
        result = self.__run_test(always)
        self.assertEqual(len(result), 7, result)

    def test_msg(self):
        result = self.__run_test(msg == MockMessage(1, "hello"))
        self.assertEqual(len(result), 1, result)

    def test_ts(self):
        result = self.__run_test(ts == 0)
        self.assertEqual(len(result), 1, result)

    def test_topic(self):
        result = self.__run_test(topic == "t1")
        self.assertEqual(len(result), 2, result)

    def test_msgtype(self):
        result = self.__run_test(msgtype == "MockMessage")
        self.assertEqual(len(result), 7, result)

        result = self.__run_test(msgtype == "FalseMessage")
        self.assertEqual(len(result), 0, result)

    def test_none_values(self):
        result = self.__run_test(msg.this.doesnt.exist == "MockMessage")
        self.assertEqual(len(result), 0, result)

        result = self.__run_test(is_none(msg.this.doesnt.exist))
        self.assertEqual(len(result), 7, result)

    def test_and(self):
        result = self.__run_test(and_(msgtype == "MockMessage", topic == "t1"))
        self.assertEqual(len(result), 2, result)

    def test_or(self):
        result = self.__run_test(or_(ts == 1, topic == "t1"))
        self.assertEqual(len(result), 3, result)

    def test_not(self):
        result = self.__run_test(not_(ts == 1))
        self.assertEqual(len(result), 6, result)

    def test_topic_match(self):
        result = self.__run_test(topic == "t1")
        self.assertEqual(len(result), 2, result)

    def test_type_match(self):
        result = self.__run_test(msgtype == "MockMessage")
        self.assertEqual(len(result), 7, result)

        result = self.__run_test(msgtype == "FalseMessage")
        self.assertEqual(len(result), 0, result)

    def test_complex_conditions(self):
        result = self.__run_test(and_(topic == "t2", msg.int_value > 2))
        self.assertEqual(len(result), 2, result)

    def test_function_calls(self):
        result = self.__run_test(msg.str_value.upper() == "HELLO")
        self.assertEqual(len(result), 4, result)

    def test_get_set_values(self):
        result = self.__run_test(
            and_(set_value("somekey", msg.str_value), get_value("somekey") == "hello")
        )
        self.assertEqual(len(result), 3, result)

    def test_has(self):
        result = self.__run_test(
            and_(has(msg.str_value, "el"), get_value("cos/contains") == "el")
        )
        self.assertEqual(len(result), 3, result)
        result = self.__run_test(
            and_(has(msg.str_value, "el"), get_value("cos/contains") == "ee")
        )
        self.assertEqual(len(result), 0, result)

    def test_regex(self):
        result = self.__run_test(
            regex(msg.str_value, r"e[lL]lo").group(0) == "ello",
        )
        self.assertEqual(len(result), 3, result)
        result = self.__run_test(
            regex(msg.str_value, r"e[lL]lo").group(0) == "eLlo",
        )
        self.assertEqual(len(result), 1, result)

    def test_concat(self):
        result = self.__run_test(
            concat(msg.str_value, "---", msg.int_value) == "hello---1"
        )
        self.assertEqual(len(result), 1, result)

    def test_coerce(self):
        result = self.__run_test(
            regex(
                msg.str_value,
                r"The value is (\d+), which is expected to be less than",
            ).group(1)
            > 111,
        )
        self.assertEqual(len(result), 1, result)

    @staticmethod
    def __run_test(condition):
        action = CollectAction()
        engine = Engine([Rule([condition], [action], {})])
        for item in simple_sequence:
            engine.consume_next(item)
        return action.collector



from inspect import signature, Parameter
from dataclasses import dataclass
from ruleengine.dsl.base_actions import upload_factory, create_moment_factory


@dataclass
class UnknownFunctionKeywordArgException(Exception):
    name: str


class ActionValidator:
    def __init__(self, action_impls):
        self.__impls = action_impls

    def upload(self, *args, **kwargs):
        def validate_arg_types(args):
            def check_is_list_of_string(name):
                value = args.get(name, [])
                if not isinstance(value, list) or any(
                    not isinstance(i, str) for i in value
                ):
                    # TODO: Make this an actual error instead of a generic error
                    raise Exception(f"{name} must be list of strings")

            check_is_list_of_string("labels")
            check_is_list_of_string("extra_files")

            if not isinstance(args.get("before", 0), int):
                # TODO: Make this an actual error instead of a generic error
                raise Exception("before must be an int")

        return self._validate_factory_func(
            upload_factory(self.__impls["upload"]),
            args,
            kwargs,
            validate_arg_types,
        )

    def create_moment(self, *args, **kwargs):
        def validate_arg_types(args):
            assign_to = args.get("assign_to", "")
            if not isinstance(assign_to, str):
                # TODO: Make this an actual error instead of a generic error
                raise Exception("assign_to must be string")

        return self._validate_factory_func(
            create_moment_factory(self.__impls["create_moment"]),
            args,
            kwargs,
            validate_arg_types,
        )

    def _validate_factory_func(self, factory, args, kwargs, param_type_check):
        sig = signature(factory)

        # We first do a manual check of the args to ensure the correct number
        # of parameters, and the right keyword args are passed. We don't use the
        # `bind` call below because it raises a TypeError, and we can't extract
        # the failure reason to report back to the user
        self._validate_signature(sig, args, kwargs)

        param_type_check(sig.bind(*args, **kwargs).arguments)

        return factory(*args, **kwargs)

    def _validate_signature(self, sig, args, kwargs):
        if len(args) > len(sig.parameters):
            raise Exception("too many arguments")

        for key in kwargs:
            if key not in sig.parameters:
                raise UnknownFunctionKeywordArgException(key)

        for i, (name, param) in enumerate(sig.parameters.items()):
            if i < len(args):
                if name in kwargs:
                    raise Exception(f"multiple values given for `{name}`")
            else:
                if name not in kwargs and param.default is Parameter.empty:
                    raise Exception(f"value not specified for parameter `{name}`")



import unittest
from collections import namedtuple

from ruleengine.dsl.action import Action
from ruleengine.dsl.base_conditions import *
from ruleengine.dsl.log_conditions import *
from ruleengine.engine import Engine, Rule, DiagnosisItem

MockMessage = namedtuple("MockMessage", "int_value str_value")
RosMockMessage = namedtuple("RosMockMessage", "msg level")
FoxgloveMockMessage = namedtuple("FoxgloveMockMessage", "message level")

simple_sequence = [
    DiagnosisItem("t1", MockMessage(1, "hello"), 0, "MockMessage"),
    DiagnosisItem("t2", MockMessage(2, "hello"), 1, "MockMessage"),
    DiagnosisItem("t1", RosMockMessage("ros log message 1", 3), 2, "rosgraph_msgs/Log"),
    DiagnosisItem("t1", RosMockMessage("ros log message 1", 4), 3, "rosgraph_msgs/Log"),
    DiagnosisItem("t1", RosMockMessage("ros log message 1", 7), 4, "MockMessage"),
    DiagnosisItem(
        "t2", FoxgloveMockMessage("foxglove log message 1", 2), 5, "foxglove_msgs/Log"
    ),
    DiagnosisItem(
        "t2", FoxgloveMockMessage("foxglove log message 2", 3), 6, "foxglove.Log"
    ),
    DiagnosisItem(
        "t2", FoxgloveMockMessage("foxglove log message 2", 7), 7, "foxglove.Log"
    ),
]


class CollectAction(Action):
    def __init__(self):
        self.collector = []

    def run(self, item, scope):
        self.collector.append(item)


class LogConditionTest(unittest.TestCase):
    def test_log_test(self):
        result = self.__run_test(log != "")
        self.assertEqual(len(result), 5, result)

        result = self.__run_test(log == "ros log message 1")
        self.assertEqual(len(result), 2, result)

        result = self.__run_test(has(log, "foxglove log message"))
        self.assertEqual(len(result), 3, result)

    def test_log_level(self):
        result = self.__run_test(log_level == LogLevel.DEBUG)
        self.assertEqual(len(result), 0, result)

        result = self.__run_test(log_level == LogLevel.INFO)
        self.assertEqual(len(result), 1, result)

        result = self.__run_test(log_level == LogLevel.WARN)
        self.assertEqual(len(result), 2, result)

        result = self.__run_test(log_level == LogLevel.UNKNOWN)
        self.assertEqual(len(result), 5, result)

    @staticmethod
    def __run_test(condition):
        action = CollectAction()
        engine = Engine([Rule([condition], [action], {})])
        for item in simple_sequence:
            engine.consume_next(item)
        return action.collector



from enum import Enum
from dataclasses import dataclass, field
from typing import Optional
from ruleengine.dsl.condition import Condition
from ruleengine.dsl.action import Action

ValidationErrorType = Enum(
    "ValidationErrorType",
    ["SYNTAX", "EMPTY", "NOT_CONDITION", "NOT_ACTION", "UNDEFINED", "UNKNOWN"],
)


@dataclass
class ValidationResult:
    success: bool

    # If not success, fill in error type and details
    error_type: Optional[ValidationErrorType] = None
    details: dict = field(default_factory=dict)

    # If success, fill in validated entity
    entity: Optional[Action | Condition] = None

    def __post_init__(self):
        if self.success:
            assert self.entity is not None
        else:
            assert self.error_type



import cProfile
from sys import argv

from ruleengine.dsl.validation.config_validator import validate_config
from ruleengine.engine import Engine, DiagnosisItem

SAMPLE_CONFIG = {
    "version": "v1",
    "rules": [
        {
            "when": [
                '"error 1" in msg',
                '"error 21444444" in msg or "asioehtnai" in msg or "aioentaihe" in msg',
                '"error 21444444" in msg or "asioehtnai" in msg or "aioentaihe" in msg',
                '"error 21444444" in msg or "asioehtnai" in msg or "aioentaihe" in msg',
                '"error 21444444" in msg or "asioehtnai" in msg or "aioentaihe" in msg',
            ],
            "actions": ["upload()"],
        }
    ],
}


def sample_action(*args, **kwargs):
    print(args, kwargs)


if __name__ == "__main__":
    res, rules = validate_config(
        SAMPLE_CONFIG, {"upload": sample_action, "create_moment": sample_action}
    )
    assert res["success"], res
    engine = Engine(rules)

    with open(argv[1]) as f:
        lines = [
            DiagnosisItem(topic=argv[1], msg=line, ts=0, msgtype="log") for line in f
        ]

    def go():
        for item in lines:
            engine.consume_next(item)

    cProfile.run("go()")



import unittest
from collections import namedtuple

from ruleengine.dsl.action import Action
from ruleengine.dsl.base_conditions import *
from ruleengine.dsl.sequence_conditions import *
from ruleengine.engine import Engine, Rule, DiagnosisItem

MockMessage = namedtuple("MockMessage", "int_value str_value")

simple_sequence = [
    DiagnosisItem("t1", MockMessage(1, "hello"), 0, "MockMessage"),
    DiagnosisItem("t1", MockMessage(2, "hello"), 0, "MockMessage"),
    DiagnosisItem("t2", MockMessage(1, "hello"), 1, "MockMessage"),
    DiagnosisItem("t2", MockMessage(2, "hello"), 1, "MockMessage"),
    DiagnosisItem("t1", MockMessage(3, "hello"), 2, "MockMessage"),
    DiagnosisItem("t1", MockMessage(3, "hello"), 3, "MockMessage"),
    DiagnosisItem("t2", MockMessage(4, "hello"), 3, "MockMessage"),
    DiagnosisItem("t2", MockMessage(4, "hello"), 3, "MockMessage"),
    DiagnosisItem("t2", MockMessage(5, "world"), 4, "MockMessage"),
    DiagnosisItem("t2", MockMessage(5, "world"), 4, "MockMessage"),
    DiagnosisItem("t2", MockMessage(4, "hello"), 5, "MockMessage"),
    DiagnosisItem("t2", MockMessage(4, "hello"), 6, "MockMessage"),
    DiagnosisItem("t2", MockMessage(4, "hello"), 7, "MockMessage"),
    DiagnosisItem("t2", MockMessage(4, "hello"), 7, "MockMessage"),
    DiagnosisItem("t2", MockMessage(4, "hello"), 9, "MockMessage"),
    DiagnosisItem("t3", MockMessage(4, "single"), 9, "MockMessage"),
]


class CollectAction(Action):
    def __init__(self):
        self.collector = []

    def run(self, item, scope):
        self.collector.append((item, scope))


def get_start_times(res):
    return [i[1]["start_time"] for i in res]


def get_trigger_times(res):
    return [i[0].ts for i in res]


class SequenceConditionTest(unittest.TestCase):
    def test_sustained_sequence(self):
        result = self.__run_test(sustained(always, msg.str_value == "hello", 2))
        self.assertEqual(get_start_times(result), [0, 5])

        result = self.__run_test(sustained(always, msg.str_value == "hello", 6))
        self.assertEqual(get_start_times(result), [])

        result = self.__run_test(sustained(topic == "t1", msg.str_value == "hello", 2))
        self.assertEqual(get_start_times(result), [0])

        result = self.__run_test(sustained(topic == "t1", msg.str_value == "hello", 6))
        self.assertEqual(get_start_times(result), [])

    def test_sequence_pattern(self):
        result = self.__run_test(
            sequential(
                and_(topic == "t1", msg.int_value == 1),
                and_(
                    topic == "t2",
                    (msg.int_value == 4),
                    set_value("somekey", msg.int_value),
                ),
                and_(topic == "t2", msg.int_value == get_value("somekey")),
                duration=4,
            )
        )
        self.assertEqual(get_start_times(result), [0])

        result = self.__run_test(
            sequential(
                and_(topic == "t1", msg.int_value == 1),
                and_(
                    topic == "t2",
                    (msg.int_value == 4),
                    set_value("somekey", msg.int_value),
                ),
                and_(topic == "t2", msg.int_value == get_value("somekey")),
                duration=2,
            )
        )
        self.assertEqual(get_start_times(result), [])

        # Overlapping sequences, and without duration
        result = self.__run_test(
            sequential(
                and_(topic == "t1", set_value("somekey", msg.int_value)),
                and_(topic == "t2", msg.int_value == get_value("somekey")),
            )
        )
        self.assertEqual(get_start_times(result), [0, 0])

    def test_sequence_timeout(self):
        result = self.__run_test(
            timeout(msg.str_value == "hello", msg.str_value == "world", duration=3)
        )
        self.assertEqual(get_start_times(result), [0, 5])

    def test_repeated(self):
        result = self.__run_test(repeated(always, 2, 5))
        self.assertEqual(get_start_times(result), [0])

        result = self.__run_test(repeated(topic == "t2", 2, 5))
        self.assertEqual(get_start_times(result), [1])

        result = self.__run_test(repeated(topic == "t2", 5, 5))
        self.assertEqual(get_start_times(result), [1])

        result = self.__run_test(repeated(topic == "t2", 2, 0.5))
        self.assertEqual(get_start_times(result), [1, 3, 4, 7])

    def test_debounce(self):
        result = self.__run_test(debounce(msg.str_value == "hello", 3))
        self.assertEqual(get_trigger_times(result), [0])

        result = self.__run_test(debounce(msg.str_value == "hello", 1.5))
        self.assertEqual(get_trigger_times(result), [0, 5, 9])

        result = self.__run_test(debounce(msg.str_value == "single", 3))
        self.assertEqual(get_trigger_times(result), [9])

    def test_throttle(self):
        result = self.__run_test(throttle(msg.str_value == "hello", 3))
        self.assertEqual(get_trigger_times(result), [0, 3, 6, 9])

    def test_any_order(self):
        result = self.__run_test(any_order(topic == "t1", topic == "t2", topic == "t3"))
        self.assertEqual(get_trigger_times(result), [9])

        result = self.__run_test(any_order(topic == "t3", topic == "t2", topic == "t1"))
        self.assertEqual(get_trigger_times(result), [9])

    @staticmethod
    def __run_test(condition):
        action = CollectAction()
        engine = Engine([Rule([condition], [action], {})])
        for item in simple_sequence:
            engine.consume_next(item)
        return action.collector



from dataclasses import dataclass
from typing import Any


@dataclass
class DiagnosisItem:
    topic: str
    msg: Any
    ts: float
    msgtype: str


@dataclass
class Rule:
    conditions: list
    actions: list
    initial_scope: dict


class Engine:
    def __init__(self, rules):
        self.__rules = rules

    def consume_next(self, item):
        for rule in self.__rules:
            for cond in rule.conditions:
                res, scope = cond.evaluate_condition_at(item, rule.initial_scope)
                if res:
                    for action in rule.actions:
                        action.run(item, scope)
                    break



import ast
import unittest

from ruleengine.dsl.validation.normalizer import normalize_expression_tree


class NormalizerTest(unittest.TestCase):
    def test_transform_boolean_ops(self):
        self._assert_equivalent("a and b and c", "and_(a, b, c)")

        self._assert_equivalent("a or b or c", "or_(a, b, c)")

        self._assert_equivalent("a or b and c", "or_(a, and_(b, c))")

        self._assert_equivalent("not a", "not_(a)")

        self._assert_equivalent(
            "not not a or b and c", "or_(not_(not_(a)), and_(b, c))"
        )

        self._assert_equivalent(
            "a in b", "(lambda arg0, arg1: and_(has(arg1, arg0)))(a, b)"
        )

        self._assert_equivalent(
            "a > 1 in b",
            "(lambda arg0, arg1, arg2: and_(arg0 > arg1, has(arg2, arg1)))(a, 1, b)",
        )

        self._assert_equivalent(
            "c or a > 1 in b and d",
            "or_(c, and_((lambda arg0, arg1, arg2: and_(arg0 > arg1, has(arg2, arg1)))(a, 1, b), d))",
        )

    def test_transform_f_strings(self):
        self._assert_equivalent(
            "f'hello {aa} world {bb} !!'",
            "func_apply(lambda arg0, arg1: f'hello {arg0} world {arg1} !!', aa, bb)",
        )

        self._assert_equivalent(
            "f'hello {1 + 1} world {s.upper()} !!'",
            "func_apply(lambda arg0, arg1: f'hello {arg0} world {arg1} !!', 1 + 1, s.upper())",
        )

        self._assert_equivalent(
            "f'hello {aa!s} world {bb!r} !!'",
            "func_apply(lambda arg0, arg1: f'hello {arg0!s} world {arg1!r} !!', aa, bb)",
        )

        self._assert_equivalent(
            "f'hello {aa:100} world {bb:5.2f} !!'",
            "func_apply(lambda arg0, arg1: f'hello {arg0:100} world {arg1:5.2f} !!', aa, bb)",
        )

        with self.assertRaises(Exception):
            normalize_expression_tree(ast.parse("f'hello {aa:{bb}}'"))

    def _assert_equivalent(self, original, normalized):
        left = normalize_expression_tree(ast.parse(original, mode="eval"))
        right = ast.parse(normalized, mode="eval")
        self.assertEqual(ast.unparse(left), ast.unparse(right))



import ast


def normalize_expression_tree(tree):
    return ast.fix_missing_locations(BooleanTransformer().visit(tree))


class BooleanTransformer(ast.NodeTransformer):
    def visit_JoinedStr(self, node):
        new_values = []
        condition_args = []
        for v in node.values:
            match v:
                case ast.Constant(value):
                    new_values.append(v)
                case ast.FormattedValue(value, conversion, format_spec):
                    if format_spec:
                        # In theory, f-strings format spec can be another
                        # f-string, but it makes our logic very clumsy, so we
                        # just don't support it for now. Check here that it's
                        # only a constant.
                        if len(format_spec.values) != 1 or not isinstance(
                            format_spec.values[0], ast.Constant
                        ):
                            raise Exception("Formatting does not support nested values")

                    new_values.append(
                        ast.FormattedValue(
                            ast.Name(f"arg{len(condition_args)}", ast.Load()),
                            conversion,
                            format_spec,
                        )
                    )
                    condition_args.append(self.generic_visit(value))
                case _:
                    raise Exception(f"Shouldn't get here: {v}")

        lambda_node = self._eval_expr(
            f"lambda {','.join(f'arg{i}' for i in range(len(condition_args)))}: ..."
        )
        lambda_node.body = ast.JoinedStr(new_values)
        return ast.Call(
            ast.Name("func_apply", ast.Load()), [lambda_node] + condition_args, []
        )

    def visit_BoolOp(self, node):
        node = self.generic_visit(node)

        match node.op:
            case ast.And():
                func_name = "and_"
            case ast.Or():
                func_name = "or_"
            case _:
                return node
        return ast.Call(ast.Name(func_name, ast.Load()), node.values, [])

    def visit_UnaryOp(self, node):
        node = self.generic_visit(node)

        match node.op:
            case ast.Not():
                return ast.Call(ast.Name("not_", ast.Load()), [node.operand], [])
            case _:
                return node

    def visit_Compare(self, node):
        node = self.generic_visit(node)

        # We need to jump through quite a few hoops to keep inline with Python's
        # semantics for comparison operators, just so we can rewrite the `in`
        # operator into our `has` function call. Suppose you have the expression
        #
        #  a op1 b op2 c
        #
        # Python treats this as
        #
        #  a op1 b and b op2 c
        #
        # Except `b` must only be evaluated once. So we package the above into
        # the following:
        #
        # (lambda arg0, arg1, arg2:
        #   and_(
        #     arg0 op1 arg1,
        #     arg1 op2 arg2))(a, b, c)
        #
        # And if one of the operators happen to be an `in` operator, we swap it
        # out with a `has` call.

        args = [node.left] + node.comparators
        param_list = ["arg" + str(i) for i in range(len(args))]
        condition_list = []
        for i, (left, right) in enumerate(zip(param_list, param_list[1:])):
            match node.ops[i]:
                case ast.In():
                    condition_list.append(self._eval_expr(f"has({right}, {left})"))
                case ast.NotIn():
                    condition_list.append(
                        self._eval_expr(f"not_(has({right}, {left}))")
                    )
                case _:
                    condition_list.append(
                        ast.Compare(
                            ast.Name(left, ast.Load()),
                            [node.ops[i]],
                            [ast.Name(right, ast.Load())],
                        )
                    )

        wrapper = self._eval_expr(f"lambda {','.join(param_list)}: ...")
        wrapper.body = ast.Call(ast.Name("and_", ast.Load()), condition_list, [])

        return ast.Call(wrapper, args, [])

    def _eval_expr(self, str_value):
        return ast.parse(str_value, mode="eval").body



import ast
from .validation_result import ValidationResult, ValidationErrorType
from .normalizer import normalize_expression_tree


def validate_expression(expr_str, injected_values):
    try:
        parsed = ast.parse(expr_str, mode="eval")
    except SyntaxError:
        return ValidationResult(False, ValidationErrorType.SYNTAX)

    for node in ast.walk(parsed):
        match node:
            case ast.Name(name, _):
                if name not in injected_values:
                    return ValidationResult(
                        False, ValidationErrorType.UNDEFINED, {"name": name}
                    )

    code = compile(normalize_expression_tree(parsed), "", mode="eval")

    return ValidationResult(True, entity=eval(code, injected_values))
